/*--------------------------------------------------------------------
  humidity2: A humidity meter using a HS1101 sensor and a 2-digit
             LED readout.

  Processor: PIC16F628A

  Discussion:
    This project is based on the Humidity1 project, substituting a
    PIC16F628A for a PIC16F684.
    
    The humidity sensor is set in an oscillator circuit generating
    a singal whose frequency is a function of the humidity, so the
    PIC needs to measure the signal frequency and convert it to
    percent humidity using a lookup table and interpolation. Note 
    that the frequency range in the circuit used varies between 
    7351 Hz (0%) and 6033 Hz (100%), so the frequency range 
    measured is quite limited.

    Timer1 is used as a 16 bit counter with external clock, the
    clock being the input signal. By turning on the timer for
    exactly 1 second, the frequency can be directly taken from
    the timer register word. There is no need to implement timer
    overflow in this application since the highest frequency measured
    is well below the 64K capability of the timer1 counter.
    
    Timer0 is used to generate 1 second intervals, at which time
    the timer1 counter is inspected,  
--------------------------------------------------------------------*/                
#include <system.h>
#include <string.h>
#include "itoa.h"

#pragma CLOCK_FREQ      4000000

// Config Bits: [00]10 0001 0001 0000
// bit 13:    Flash Program Memory Code Protection, 1=protection off
// bit 12-9:  unimplemented, set to 0
// bit 8:     Data Code Protection, 1=protection off
// bit 7:     LVP: Low-Voltage Programming Enable, 0=use HV programming
// bit 6:     BOREN: Brown-out Reset Enable, 0=BOR disabled
// bit 5:     MCLRE  0 = RA5/!MCLR pin function is RA5
// bit 4:     1 (part of FOSC bits)     
// bit 3:     ^PWRTE 0 = PWRT enabled
// bit 2:     WDTE   0 = watchdog timer disabled
// bits [4]1-0:  FOSC [1]00 = INTOSC oscillator: I/O function on RA6/OSC2/CLKOUT pin, 
//							  I/O function on RA7/OSC1/CLKIN

#pragma DATA _CONFIG, 0x2110

#define TIMER0_START_CNT	56		// tweak this value to calibrate
#define MIN_RH				0
#define MAX_RH				99

// Global Variables

static unsigned int  uiTimer1Ctr;

// Functions

static unsigned short freq2rh( unsigned short uiFreq );
static void initCPU( void );
static void ledDigit( char cDigit );
static void ledOutput( char cNumber );

/*---------------------------------------------------------------------
  main()
---------------------------------------------------------------------*/
void main()
{
  char buffer[4];
  char cDigit, cNumber, cLen, cStrt;
  int  iRh;
  
  initCPU();                	// initialize hardware
  uiTimer1Ctr = 0;
  set_bit( t1con, TMR1ON );		// start timer1 counter

  // The mainline job is to convert timer1 counts to ascii and display
  // the LED digits. Everything else is managed by timer0 and timer1
  // interrupt handlers.

  while ( 1 )
  {
    // Convert the frequency reading (7351..6033 Hz) to relative
    // humidity.
    
    iRh = freq2rh( uiTimer1Ctr );
    
    // The number in buffer has no leading zeroes. We want to display
    // the digits right-justified. To do this we skip the high-order
    // digit when there is nothing to display.
   
    itoa( iRh, buffer );
    cLen = strlen( buffer );
    cStrt = 2 - cLen;
    
    for ( cDigit=0; cDigit<cLen; ++cDigit, ++cStrt )
    {
      cNumber = buffer[cDigit];
      ledDigit( 0 );					// turn both digits off
      ledOutput( cNumber );				// output a number
      ledDigit( cStrt + 1 );			// turn the digit on
      delay_ms( 2 );
    }  
  }
}

/*---------------------------------------------------------------------
  interrupt()
  
  Handles the timer interrupts.
---------------------------------------------------------------------*/
void interrupt( void )
{
  static unsigned short t0ctr=0;
  static unsigned char diag=0;
  
  if ( test_bit( intcon, T0IF ) )
  {
    // The timer0 interrupt occurs ( Fosc/4 * prescale * t0 counts ). When 
    // Fosc/4 is 1 microsecond and prescale is 1:2, the timer increments
    // every 2 microseconds. If we count 250 times, we get an interrupt every 
    // 500 microseconds. Since we are timing one second intervals, we need to
    // count to 2000.
    ++t0ctr;
    if ( t0ctr == 2000 )
    {
      clear_bit( t1con, TMR1ON );		// stop timer1 counter
      MAKESHORT( uiTimer1Ctr, tmr1l, tmr1h );

      tmr1l = tmr1h = 0;
      set_bit( t1con, TMR1ON );		// start timer1 counter
      
      tmr0 = TIMER0_START_CNT;		// set timer0 starting value
      t0ctr = 0;					// reset the interrupt counter
      
	  // Add some code to turn port RB7 on and off every second.
      if ( diag )
      {
        diag = 0;
        clear_bit( portb, 7 );
      }
      else
      {
        diag = 1;
        set_bit( portb, 7 );
      }    
    }
    
    clear_bit( intcon, T0IF );		// reset timer0 overflow interrupt flag
    return;
  }
}

/*---------------------------------------------------------------------
  initCPU()
---------------------------------------------------------------------*/
static void initCPU()
{
  // PORTB connections as follows:
  //  RB0 - (out) LED data 'A'
  //  RB1 - (out) LED data 'B'
  //  RB2 - (out) LED data 'C'
  //  RB3 - (out) LED data 'D'
  //  RB4 - (out) LED digit select 'DIG1'
  //  RB5 - (out) LED digit select 'DIG2'
  //  RB6 - (in)  signal input to timer1 counter
  //  RB7 - (out) diagnostic LED 

  trisb = 01000000b;            // 1=input, 0=output
  portb = 0;

  // PORTA is not used.

  trisa = 00000000b;			// 1=input, 0=output
  porta = 0;

  // Timer0 initialization. Use Timer0 to measure one second intervals.
  tmr0 = TIMER0_START_CNT;		// set timer0 starting value
  option_reg = 11010000b;		// internal clock; 1:2 prescale;
  set_bit( intcon, T0IE );		// enable timer0 interrupts
  
  // Timer1 initialization. Use Timer1 as 16-bit counter with
  // input on RA5/T1CKI (no interrupts).

  tmr1l = tmr1h = 0;
  t1con = 00000110b;      	// 1:1 prescale; LP osc off; no sync; ext. clock; stopped

  // Enable interrupts for Timer0

  set_bit( intcon, GIE );       // enable global interrupts
  set_bit( intcon, PEIE );      // enable peripheral interrupts
}

/*---------------------------------------------------------------------
  ledDigit(...)
  
  Selects a digit of the led display. Ports RB4-RB5 are connected
  to DIG1 and DIG2 of the 2-digit LED display.
  
  Input: 0 = deactivate both digits
         1 = activate digit 1
         2 = activate digit 2
---------------------------------------------------------------------*/
static void ledDigit( char cDigit )
{
  if ( cDigit == 1 )
  {
    // Activate digit 1
    set_bit( portb, 5 );
    clear_bit( portb, 4 );
  }
  else
    if ( cDigit == 2 )
    {
      // Activate digit 2
      set_bit( portb, 4 );
      clear_bit( portb, 5 );
    }
    else
    {
      set_bit( portb, 4 );
      set_bit( portb, 5 );
    }
}

/*---------------------------------------------------------------------
  ledOutput(...)
  
  Outputs a number to the led display. Ports RB0..RB3 are connected
  to the A-B-C-D inputs of a 74LS47 which decodes the number and
  activates the corresponding segments of the bank of LEDs. Note that
  A is the low-order input and D is the high order input to the '47.
  
  Input: a single numeric digit, range 0..9.
---------------------------------------------------------------------*/
static void ledOutput( char cNumber )
{
  // Output the digit as BCD on RB0..RB3

  if ( test_bit( cNumber, 0 ) )
    set_bit( portb, 0 );
  else
    clear_bit( portb, 0 );

  if ( test_bit( cNumber, 1 ) )
    set_bit( portb, 1 );
  else
    clear_bit( portb, 1 );

  if ( test_bit( cNumber, 2 ) )
    set_bit( portb, 2 );
  else
    clear_bit( portb, 2 );

  if ( test_bit( cNumber, 3 ) )
    set_bit( portb, 3 );
  else
    clear_bit( portb, 3 );
}

/*---------------------------------------------------------------------
  freq2rh(...)

  Converts the frequency to relative humidity, using linear interpolation.

  Linear interpolation takes two data points, say (Xa,Ya) and (Xb,Yb),
  and the interpolant (X,Y) is given by:

           (X - Xa) (Yb - Ya)
  Y = Ya + ------------------
               (Xb - Xa)

  In this case X is frequency and Y is relative humidity.

  Input: frequency 7351..6033 Hz
---------------------------------------------------------------------*/
static unsigned short freq2rh( unsigned short uiFreq )
{
  static short sFreq[11] =
  {
    7351, 7224, 7100, 6976, 6853, 6728, 6600, 6468, 6330, 6186, 6033
  };
  static short sRh[11] =
  {
    MIN_RH, 10, 20, 30, 40, 50, 60, 70, 80, 90, MAX_RH
  };

  char  ii;
  unsigned usRH;
  short sTerm1, sTerm2, sTerm3;

  // Check for boundary values
  if ( uiFreq <= (unsigned short) sFreq[10] ) return MAX_RH;
  if ( uiFreq >= (unsigned short) sFreq[0] )  return MIN_RH;

  for ( ii=0; ii<11; ++ii )
  {
    if ( uiFreq >= (unsigned) sFreq[ii] )
      break;
  }

  if ( uiFreq == (unsigned) sFreq[ii] )
    return (unsigned) sRh[ii];          // no interpolation necessary

  sTerm1 = (short) uiFreq - sFreq[ii-1];
  sTerm2 = sRh[ii] - sRh[ii-1];
  sTerm3 = sFreq[ii] - sFreq[ii-1];
  usRH = (unsigned) ( sRh[ii-1] + ( ( sTerm1 * sTerm2 ) / sTerm3 ) );

  return usRH;
}
